__all__ = ('SolarNode', )

from scarletio import (
    Future, RichAttributeErrorBaseType, Task, from_json, future_or_timeout, repeat_timeout, sleep, to_json,
    write_exception_async
)
from scarletio.web_common import ConnectionClosed, InvalidHandshake, WebSocketProtocolError

from ...discord.channel import VoiceRegion
from ...discord.core import KOKORO

from .constants import (
    HEADER_AUTHORIZATION, HEADER_CLIENT_NAME, HEADER_RESUME_KEY, HEADER_SHARD_COUNT, HEADER_USER_ID,
    LAVALINK_KEY_EVENT_TYPE, LAVALINK_KEY_GATEWAY_OPERATION_EVENT, LAVALINK_KEY_GATEWAY_OPERATION_PLAYER_UPDATE,
    LAVALINK_KEY_GATEWAY_OPERATION_STATS, LAVALINK_KEY_GUILD_ID, LAVALINK_KEY_NODE_OPERATION,
    LAVALINK_KEY_NODE_OPERATION_SET_RESUME_KEY, LAVALINK_KEY_NODE_RESUME_KEY, LAVALINK_KEY_PLAYER_STATE
)
from .exceptions import SolarAuthenticationError
from .parsers import PARSERS
from .stats import Stats


EXTENSION_USER_AGENT = 'hata.ext.solarlink'
DEFAULT_PENALTY = 9e30

class SolarNode(RichAttributeErrorBaseType):
    """
    Represents a Node connection with Lavalink.
    
    Attributes
    ----------
    _host : `str`
        Host of a lavalink node to connect to.
    _password : `str`
        The password used for authentication.
    _port : `int`
        Port of a lavalink node to connect to.
    _resume_key : `None`, `str`
        A resume key to resume websocket connection with lavalink.
    client : ``Client``
        The parent client of the node.
    players : `set` of ``SolarPlayerBase``
        The players using this node.
    reconnect_attempts : `int`
        How much times the gateway should try to reconnect before erroring out.
        
        Defaults to `3`. If defined as a negative number, the node will try to reconnect infinitely.
    
    region : `None`, ``VoiceRegion``
        The respective voice region of the node's players.
    stats : `None`, ``Stats``
        The statistics of the node.
    websocket : `None`, ``WebSocketClient``
        The connected websocket.
    """
    __slots__ = (
        '_host', '_password', '_port', '_resume_key', 'client', 'players', 'reconnect_attempts', 'region', 'stats',
        'websocket'
    )
    
    def __new__(cls, client, host, port, password, region, resume_key, reconnect_attempts):
        """
        Creates a new node connection with the given parameters.
        
        Parameters
        ----------
        client : ``Client``
            The parent client.
        host : `str`
            The host of the lavalink node to connect to.
        port : `int`
            The port of the lavalink node to connect to.
        password : `str`
            The password used for authentication.
        region : `None`, ``VoiceRegion``
            The respective voice region of the node's players.
        resume_key : `None`, `str`
            A resume key to resume websocket connection with lavalink.
        reconnect_attempts : `int`
            How much times the gateway should try to reconnect before erroring out.
        
        Raises
        ------
        TypeError
            - If `host` is not `str`.
            - If `port` is not `int`.
            - If `password is not `str`.
            - If `region` is neither `None`, nor ``VoiceRegion``.
            - If `resume_key` is neither `None`, nor `str`.
            - If `reconnect_attempts` is not `int`.
        """
        if type(host) is str:
            pass
        elif isinstance(host, str):
            host = str(host)
        else:
            raise TypeError(
                f'`host` can be `str`, got {host.__class__.__name__}; {host!r}.'
            )
        
        if type(port) is int:
            pass
        elif isinstance(port, int):
            port = int(port)
        else:
            raise TypeError(
                f'`port` can be `int`, got {port.__class__.__name__}; {port!r}.'
            )
        
        if type(password) is str:
            pass
        elif isinstance(password, str):
            password = str(password)
        else:
            raise TypeError(
                f'`password` can be `str`, got {password.__class__.__name__}; {password!r}.'
            )
        
        if (region is not None) and (not isinstance(region, VoiceRegion)):
            raise TypeError(
                f'`region` can be `None`, `{VoiceRegion.__name__}`, got '
                f'{region.__class__.__name__}; {region!r}.'
            )
        
        if (resume_key is None) or (type(resume_key) is str):
            pass
        elif isinstance(resume_key, str):
            resume_key = str(resume_key)
        else:
            raise TypeError(
                f'`resume_key` can be `None`, `str`, got '
                f'{resume_key.__class__.__name__}; {resume_key!r}.'
            )
        
        if type(reconnect_attempts) is int:
            pass
        elif isinstance(reconnect_attempts, int):
            reconnect_attempts = int(reconnect_attempts)
        else:
            raise TypeError(
                f'`reconnect_attempts` can be `int`, got '
                f'{reconnect_attempts.__class__.__name__}; {reconnect_attempts!r}.'
            )
        
        self = object.__new__(cls)
        
        self.client = client
        self._host = host
        self._port = port
        self._password = password
        self.region = region
        self.stats = None
        self.players = set()
        self.websocket = None
        self._resume_key = resume_key
        self.reconnect_attempts = reconnect_attempts
        return self
    
    
    @property
    def available(self):
        """
        Returns whether the node is available for requests.
        
        Returns
        -------
        available : `bool`
        """
        websocket = self.websocket
        if websocket is None:
            available = False
        else:
            available = websocket.open
        
        return available
    
    
    @property
    def penalty(self):
        """
        Returns the load-balancing penalty for this node.
        
        Returns
        -------
        penalty : `float`
        """
        if self.available:
            stats = self.stats
            if (stats is None):
                penalty = DEFAULT_PENALTY
            else:
                penalty = stats.total_penalty
        else:
            penalty = DEFAULT_PENALTY
        
        return penalty
    
    
    async def _send(self, data):
        """
        Sends the passed data to the node via the websocket connection.
        
        Parameters
        ----------
        data : `dict`
            The dict to send to Lavalink.
        """
        data = to_json(data)
        
        websocket = self.websocket
        if (websocket is not None):
            await websocket.send(data)
    
    
    async def _connect(self):
        """
        Connects to a lavalink node.
        
        This method is a coroutine.
        
        Raises
        ------
        ConnectionError
        OSError
        ValueError
        ConnectionClosed
        InvalidHandshake
        WebSocketProtocolError
        """
        client = self.client
        
        shard_count = client.shard_count
        if not shard_count:
            shard_count = 1
        
        headers = {
            HEADER_AUTHORIZATION: self._password,
            HEADER_USER_ID : str(client.id),
            HEADER_CLIENT_NAME: EXTENSION_USER_AGENT,
            HEADER_SHARD_COUNT: str(shard_count),
        }
        
        resume_key = self._resume_key
        if (resume_key is not None):
            headers[HEADER_RESUME_KEY] = resume_key
        
        while True:
            websocket = self.websocket
            if (websocket is not None) and (not websocket.closed):
                await websocket.close(4000)
                self.websocket = None
            
            self.websocket = await self.client.http.connect_websocket(
                f'ws://{self._host}:{self._port}',
                headers = headers,
            )
            
            try:
                await self.websocket.ensure_open()
            except ConnectionClosed:
                continue
            
            break
        
        if (resume_key is not None):
            # Resume timeout defaults to `60` seconds, we will keep it.
            await self._send({
                LAVALINK_KEY_NODE_OPERATION: LAVALINK_KEY_NODE_OPERATION_SET_RESUME_KEY,
                LAVALINK_KEY_NODE_RESUME_KEY: resume_key,
            })
        
        await self.client.solarlink._node_connected(self)
    
    
    async def run(self, waiter=None):
        """
        Keeps the node's gateway open.
        
        This method is a coroutine.
        """
        client = self.client
        reconnect_attempts = self.reconnect_attempts
        
        try:
            while True:
                try:
                    task = Task(self._connect(), KOKORO)
                    future_or_timeout(task, 30.0)
                    await task
                    
                    if (waiter is not None):
                        waiter.set_result_if_pending(True)
                        waiter = None
                    
                    while True:
                        try:
                            with repeat_timeout(60.0) as loop:
                                for _ in loop:
                                    should_reconnect = await self._poll_event()
                                    if should_reconnect:
                                        break
                        except TimeoutError:
                            # timeout, no internet probably; try to reconnect
                            
                            # Make sure we are still online
                            if not client.running:
                                return
                            
                            # Try to reconnect later.
                            await sleep(5.0, KOKORO)
                            break
                        
                        if not client.running:
                            return
                        
                        reconnect_attempts = self.reconnect_attempts
                        task = Task(self._connect(), KOKORO)
                        future_or_timeout(task, 30.0)
                        await task
                
                
                except (OSError, TimeoutError, ConnectionError, ConnectionClosed, WebSocketProtocolError,
                        InvalidHandshake, ValueError) as err:
                    
                    if not client.running:
                        return
                    
                    reconnect_attempts -= 1
                    
                    if isinstance(err, InvalidHandshake):
                        if err.response.status in (401, 403):
                            raise SolarAuthenticationError(self, err.response) from None
                        
                        sleep_amount = 0.0
                    
                    elif isinstance(err, ConnectionClosed):
                        sleep_amount = 0.0
                    
                    elif isinstance(err, TimeoutError):
                        sleep_amount = 0.0
                    
                    elif isinstance(err, ConnectionError): # no internet
                        sleep_amount = 5.0
                    
                    else:
                        sleep_amount = 1.0
                    
                    if not reconnect_attempts:
                        await write_exception_async(
                            err,
                            [
                                'Failed to connect to lavalink node ',
                                repr(self),
                                '.run:\n',
                            ],
                            loop = KOKORO,
                        )
                        
                        if (waiter is not None):
                            waiter.set_result_if_pending(False)
                            waiter = None
                        
                        return
                    
                    if sleep_amount:
                        await sleep(sleep_amount, KOKORO)
                    
                    continue
                
        
        except GeneratorExit:
            if (waiter is not None):
                waiter.set_result_if_pending(False)
                waiter = None
            
            raise
        
        except BaseException as err:
            if (waiter is not None):
                waiter.set_exception_if_pending(err)
                waiter = None
            
            raise
        finally:
            await client.solarlink._node_disconnected(self)
            
            if (waiter is not None):
                waiter.set_result_if_pending(False)
                waiter = None
    
    
    async def _poll_event(self):
        """
        Polls an event from the gateway and tries to handle it.
        
        This method is a coroutine.
        
        Returns
        -------
        should_reconnect : `bool`
        
        Raises
        ------
        ConnectionClosed
            If the websocket connection closed.
        """
        websocket = self.websocket
        if websocket is None:
            return True
        
        try:
            message = await websocket.receive()
        except ConnectionClosed as err:
            raise err
        
        return (await self._received_message(message))
    
    
    async def _received_message(self, message):
        """
        Processes a message received from a lavalink node.
        
        Parameters
        ----------
        message : `str`
            the received message.
        
        Returns
        -------
        should_reconnect : `bool`
        """
        message = from_json(message)
        
        operation = message[LAVALINK_KEY_NODE_OPERATION]
        
        if operation == LAVALINK_KEY_GATEWAY_OPERATION_STATS:
            self.stats = Stats(message)
            return False
        
        if operation == LAVALINK_KEY_GATEWAY_OPERATION_PLAYER_UPDATE:
            guild_id = int(message[LAVALINK_KEY_GUILD_ID])
            
            try:
                player = self.client.solarlink.players[guild_id]
            except KeyError:
                return False
            
            player._update_state(message[LAVALINK_KEY_PLAYER_STATE])
            return False
        
        elif operation == LAVALINK_KEY_GATEWAY_OPERATION_EVENT:
            event = message[LAVALINK_KEY_EVENT_TYPE]
            client = self.client
            
            try:
                parser = PARSERS[event]
            except KeyError:
                Task(
                    client.events.error(
                        client,
                        f'{self.__class__.__name__}._received_message',
                        f'Unknown dispatch event {event}\nData: {message!r}',
                    ),
                    KOKORO,
                )
                return False
            
            try:
                parser(self.client, message)
            except BaseException as err:
                Task(client.events.error(client, event, err), KOKORO)
            
            return False
        
        client = self.client
        Task(
            client.events.error(
                client,
                f'{self.__class__.__name__}._received_message',
                f'Unknown operation {operation}\nData: {message!r}',
            ),
            KOKORO,
        )
        return False
    
    
    async def start(self):
        """
        Starts the lavalink node.
        
        This method is a coroutine.
        
        Raises
        ------
        BaseException
            Any exception raised when trying to connect.
        """
        waiter = Future(KOKORO)
        Task(self.run(waiter=waiter), KOKORO)
        return await waiter
    
    
    async def close(self):
        """
        Closes it's ``.websocket`` with close code of `1000`.
        
        This method is a coroutine.
        """
        websocket = self.websocket
        if websocket is None:
            return
        
        self.websocket = None
        await websocket.close(1000)
    
    
    def __repr__(self):
        """Returns the node's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' client=')
        repr_parts.append(repr(self.client))
        
        repr_parts.append(', host=')
        repr_parts.append(repr(self._host))
        
        repr_parts.append(', port=')
        repr_parts.append(repr(self._port))
        
        region = self.region
        if (region is not None):
            repr_parts.append(', region=')
            repr_parts.append(region.name)
        
        resume_key = self._resume_key
        if (resume_key is not None):
            repr_parts.append(', resume_key=')
            repr_parts.append(repr(resume_key))
        
        repr_parts.append(', reconnect_attempts=')
        repr_parts.append(repr(self.reconnect_attempts))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
        





__all__ = ('AutoModerationKeywordPresetType', )

from ...bases import Preinstance as P, PreinstancedBase


class AutoModerationKeywordPresetType(PreinstancedBase):
    """
    Represents an auto moderation keyword preset type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the auto moderation keyword preset type.
    name : `str`
        The default name of the auto moderation keyword preset type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AutoModerationKeywordPresetType``) items
        Stores the predefined auto moderation keyword preset types. This container is accessed when translating a
        Discord side identifier of a auto moderation keyword preset type. The identifier value is used as a key to
        get it's wrapper side representation.
    VALUE_TYPE : `type` = `str`
        The auto moderation keyword preset types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the auto moderation keyword preset types.
    
    Every predefined auto moderation keyword preset type is also stored as a class attribute:
    
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | Class attribute name  | Name                  | Value     | Description                               |
    +=======================+=======================+===========+===========================================+
    | none                  | none                  | 0         | N/A                                       |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | cursing               | cursing               | 1         | Swearing or cursing.                      |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | sexually_suggestive   | sexually suggestive   | 2         | Sexually explicit behavior or activity.   |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | slur                  | slur                  | 3         | Personal insult or hate speech.           |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    # predefined
    none = P(0, 'none')
    cursing = P(1, 'cursing')
    sexually_suggestive = P(2, 'sexually suggestive')
    slur = P(3, 'slur')



__all__ = ()

from scarletio import Compound

from ...auto_moderation import AutoModerationRule
from ...http import DiscordHTTPClient

from ..request_helpers import (
    get_auto_moderation_rule_and_guild_id_and_id, get_auto_moderation_rule_guild_id_and_id, get_guild_id
)


class ClientCompoundAutoModerationEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    async def auto_moderation_rule_get(self, auto_moderation_rule):
        """
        Requests the specified auto moderation rule.
        
        This method is a coroutine.
        
        Parameters
        ----------
        auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
            The auto moderation rule to get, or a `guild-id`, `rule-id` pair representing it.
        
        Returns
        -------
        auto_moderation_rule : ``AutoModerationRule``
        
        Raises
        ------
        TypeError
            If `auto_moderation_rule` was not given neither as ``AutoModerationRule`` nor `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        auto_moderation_rule, guild_id, auto_moderation_rule_id = get_auto_moderation_rule_and_guild_id_and_id(
            auto_moderation_rule
        )
        
        auto_moderation_rule_data = await self.http.auto_moderation_rule_get(guild_id, auto_moderation_rule_id)
        
        if (auto_moderation_rule is None):
            auto_moderation_rule = AutoModerationRule.from_data(auto_moderation_rule_data)
        
        else:
            auto_moderation_rule._update_attributes(auto_moderation_rule_data)
        
        return auto_moderation_rule
    
    
    async def auto_moderation_rule_get_all(self, guild):
        """
        Requests all the auto moderation rules of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild or it's identifier.
        
        Returns
        -------
        auto_moderation_rules : `list` of``AutoModerationRule``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        auto_moderation_rule_datas = await self.http.auto_moderation_rule_get_all(guild_id)
        
        return [
            AutoModerationRule.from_data(auto_moderation_rule_data)
            for auto_moderation_rule_data in auto_moderation_rule_datas
        ]
    
    
    async def auto_moderation_rule_create(self, guild, auto_moderation_rule, *, reason = None):
        """
        Creates an auto moderation rule at the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild or it's identifier.
        auto_moderation_rule : ``AutoModerationRule``
            The auto moderation rule to create one like.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Returns
        -------
        auto_moderation_rule : ``AutoModerationRule``
            The created auto moderation rule.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `auto_moderation_rule` was not given as ``AutoModerationRule``.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        if not isinstance(auto_moderation_rule, AutoModerationRule):
            raise TypeError(
                f'`auto_moderation_rule` can be {AutoModerationRule.__name__}, '
                f'got {auto_moderation_rule.__class__.__name__}; {auto_moderation_rule!r}.'
            )
        
        data = auto_moderation_rule.to_data()
        
        auto_moderation_rule_data = await self.http.auto_moderation_rule_create(guild_id, data, reason)
        
        return AutoModerationRule.from_data(auto_moderation_rule_data)
    
    
    async def auto_moderation_rule_edit(self, old_auto_moderation_rule, new_auto_moderation_rule, *, reason = None):
        """
        Edits the specified auto moderation rule
        
        This method is a coroutine.
        
        Parameters
        ----------
        old_auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
            The auto moderation rule to edit, or a `guild-id`, `rule-id` pair representing it.
        new_auto_moderation_rule : ``AutoModerationRule``
            The auto moderation rule to edit the current to.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Returns
        -------
        auto_moderation_rule : ``AutoModerationRule``
        
        Raises
        ------
        TypeError
            - If `old_auto_moderation_rule` was not given neither as ``AutoModerationRule`` nor `tuple` (`int`, `int`).
            - If `new_auto_moderation_rule` was not given as ``AutoModerationRule``.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, auto_moderation_rule_id = get_auto_moderation_rule_guild_id_and_id(old_auto_moderation_rule)
    
        if not isinstance(new_auto_moderation_rule, AutoModerationRule):
            raise TypeError(
                f'`new_auto_moderation_rule` can be {AutoModerationRule.__name__}, '
                f'got {new_auto_moderation_rule.__class__.__name__}; {new_auto_moderation_rule!r}.'
            )
        
        data = new_auto_moderation_rule.to_data()
        
        await self.http.auto_moderation_rule_edit(guild_id, auto_moderation_rule_id, data, reason)
    
    
    async def auto_moderation_rule_delete(self, auto_moderation_rule, *, reason = None):
        """
        Deletes the specified auto moderation rule.
        
        This method is a coroutine.
        
        Parameters
        ----------
        auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
            The auto moderation rule to delete, or a `guild-id`, `rule-id` pair representing it.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `auto_moderation_rule` was not given neither as ``AutoModerationRule`` nor `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, auto_moderation_rule_id = get_auto_moderation_rule_guild_id_and_id(auto_moderation_rule)
        await self.http.auto_moderation_rule_delete(guild_id, auto_moderation_rule_id, reason)



__all__ = ('Integration', )

import warnings

from scarletio import copy_docs, export

from ...bases import DiscordEntity
from ...core import INTEGRATIONS
from ...preconverters import preconvert_snowflake
from ...role import create_partial_role_from_id
from ...user import ZEROUSER

from ..integration_metadata import IntegrationMetadataBase, IntegrationMetadataSubscription

from .fields import (
    parse_enabled, parse_name, parse_type, parse_user, put_enabled_into, put_name_into, put_type_into, put_user_into,
    validate_enabled, validate_name, validate_type, validate_user
)
from .preinstanced import IntegrationType


@export
class Integration(DiscordEntity, immortal = True):
    """
    Represents a Discord Integration.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the integration.
    enabled : `bool`
        Whether this integration is enabled.
    name : `str`
        The name of the integration.
    metadata : ``IntegrationMetadataBase``
        type specific information of the integration.
    type : ``IntegrationType``
        The type of the integration.
    user : ``ClientUserBase``
        User for who the integration is.
    """
    __slots__ = ('enabled', 'name', 'metadata', 'type', 'user')
    
    def __new__(
        cls,
        *,
        integration_type = ...,
        enabled = ...,
        name = ...,
        user = ...,
        **keyword_parameters,
    ):
        """
        Creates a new integration from the given keyword parameters.
        
        Parameters
        ----------
        integration_type : `str`, ``IntegrationType``, Optional (Keyword only)
            The integration's type.
        
        enabled : `bool`, Optional (Keyword only)
            Whether this integration is enabled.
        
        name : `str`, Optional (Keyword only)
            The name of the integration.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            User for who the integration is.
        
        **keyword_parameters : Keyword parameters
            keyword parameters to set the integration's attributes as.
        
        Other Parameters
        ----------------
        account : ``IntegrationAccount``, ``ClientUserBase``, Optional (Keyword only)
            The integration's respective account. If the integration type is `'discord'`, then set as a discord user
            itself.
        
        application : ``IntegrationApplication``, Optional (Keyword only)
            The application of the integration if applicable.
        
        expire_behavior : `int`, ``IntegrationExpireBehavior``, Optional (Keyword only)
            The behavior of expiring subscription.
        
        expire_grace_period : `int`, Optional (Keyword only)
            The grace period in days for expiring subscribers.
        
        role_id : `int`, Optional (Keyword only)
            The role's identifier what the integration uses for subscribers.
        
        scopes : `None`, `iterable` of (`str`, ``Oauth2Scope``) items, Optional (Keyword only)
            The scopes the application was authorized with.
        
        subscriber_count : `int`, Optional (Keyword only)
            How many subscribers the integration has. 
        
        synced_at : `None`, `datetime`, Optional (Keyword only)
            When the integration was last synced.
        
        syncing : `bool`, Optional (Keyword only)
            Whether the integration syncing.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - Parameter type incorrect.
            - Extra or unused parameters.
        ValueError
            - Parameter value incorrect.
        """
        # integration_type
        if integration_type is ...:
            integration_type = IntegrationType.none
        else:
            integration_type = validate_type(integration_type)
        
        # enabled
        if enabled is ...:
            enabled = False
        else:
            enabled = validate_enabled(enabled)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        metadata = integration_type.metadata_type(keyword_parameters)
            
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        self = object.__new__(cls)
        self.id = 0
        self.enabled = enabled
        self.name = name
        self.metadata = metadata
        self.type = integration_type
        self.user = user
        return self
    
    
    @classmethod
    def precreate(cls, integration_id, *, integration_type = ..., **keyword_parameters):
        """
        Parameters
        ----------
        integration_id : `int`
            The integration's identifier.
        
        integration_type : `str`, ``IntegrationType``, Optional (Keyword only)
            The integration's type.
        
        **keyword_parameters : Keyword parameters
            keyword parameters to set the integration's attributes as.
        
        Other Parameters
        ----------------
        account : ``IntegrationAccount``, ``ClientUserBase``, Optional (Keyword only)
            The integration's respective account. If the integration type is `'discord'`, then set as a discord user
            itself.
        
        application : ``IntegrationApplication``, Optional (Keyword only)
            The application of the integration if applicable.
        
        enabled : `bool`, Optional (Keyword only)
            Whether this integration is enabled.
        
        expire_behavior : `int`, ``IntegrationExpireBehavior``, Optional (Keyword only)
            The behavior of expiring subscription.
        
        expire_grace_period : `int`, Optional (Keyword only)
            The grace period in days for expiring subscribers.
        
        name : `str`, Optional (Keyword only)
            The name of the integration.
        
        role_id : `int`, Optional (Keyword only)
            The role's identifier what the integration uses for subscribers.
        
        scopes : `None`, `iterable` of (`str`, ``Oauth2Scope``) items, Optional (Keyword only)
            The scopes the application was authorized with.
        
        subscriber_count : `int`, Optional (Keyword only)
            How many subscribers the integration has. 
        
        synced_at : `None`, `datetime`, Optional (Keyword only)
            When the integration was last synced.
        
        syncing : `bool`, Optional (Keyword only)
            Whether the integration syncing.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            User for who the integration is.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - Parameter type incorrect.
            - Extra or unused parameters.
        ValueError
            - Parameter value incorrect.
        """
        integration_id = preconvert_snowflake(integration_id, 'integration_id')
        
        if (integration_type is not ...) or keyword_parameters:
            processable = []
            
            if integration_type is ...:
                integration_type = IntegrationType.none
            else:
                integration_type = validate_type(integration_type)
            processable.append(('type', integration_type))
            
            # enabled
            try:
                enabled = keyword_parameters.pop('enabled')
            except KeyError:
                pass
            else:
                enabled = validate_enabled(enabled)
                processable.append(('enabled', enabled))
            
            # name
            try:
                name = keyword_parameters.pop('name')
            except KeyError:
                pass
            else:
                name = validate_name(name)
                processable.append(('name', name))
            
            # metadata
            metadata = integration_type.metadata_type(keyword_parameters)
            processable.append(('metadata', metadata))
            
            # user
            try:
                user = keyword_parameters.pop('user')
            except KeyError:
                pass
            else:
                user = validate_user(user)
                processable.append(('user', user))
            
            if keyword_parameters:
                raise TypeError(
                    f'Extra or unused keyword parameters: {keyword_parameters!r}.'
                )
        
        else:
            processable = None
        
        try:
            self = INTEGRATIONS[integration_id]
        except KeyError:
            self = cls._create_empty(integration_id)
            INTEGRATIONS[integration_id] = self
        
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, integration_id):
        """
        Creates a new integration with it's default attributes set.
        
        Parameters
        ----------
        integration_id : `int`
            The integration's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = integration_id
        self.enabled = True
        self.name = ''
        self.metadata = IntegrationMetadataBase._create_empty()
        self.type = IntegrationType.none
        self.user = ZEROUSER
        return self
    
    
    @classmethod
    def _create_with_role(cls, integration_id, role_id):
        """
        Creates a new integration with the given `role`.
        
        Parameters
        ----------
        integration_id : `int`
            The integration's identifier.
        role_id : `int`
            The role's identifier to create the integration with.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = cls._create_empty(integration_id)
        
        if role_id:
            self.metadata = IntegrationMetadataSubscription.from_role(role_id)
        
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new integration object with the given data. If the integration already exists, then updates and
        returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Integration data received from Discord.
        
        Returns
        -------
        integration : `instance<cls>`
        """
        integration_id = int(data['id'])
        
        try:
            self = INTEGRATIONS[integration_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = integration_id
            self._set_attributes(data)
            
            INTEGRATIONS[integration_id] = self
        
        else:
            self._set_attributes(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the integration into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default fields should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        # metadata
        data = self.metadata.to_data(defaults = defaults)
        
        # id
        if include_internals:
            data['id'] = str(self.id)
        
        # enabled
        put_enabled_into(self.enabled, data, defaults)
        
        # name
        put_name_into(self.name, data, defaults)
        
        # type
        put_type_into(self.type, data, defaults)
        
        # user
        if include_internals:
            put_user_into(self.user, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data):
        """
        Sets the attributes of the integration from the given `data`.
        
        > `.id` field should be set already.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Integration data received from Discord.
        """
        integration_type = parse_type(data)
        
        self.enabled = parse_enabled(data)
        self.name = parse_name(data)
        self.metadata = integration_type.metadata_type.from_data(data)
        self.type = integration_type
        self.user = parse_user(data)
        
        return self
    
    
    @property
    def partial(self):
        """
        Returns whether the integration is partial.
        
        Returns
        -------
        partial : `bool`
        """
        if not self.id:
            return True
        
        if self.type is IntegrationType.none:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the integration's representation."""
        repr_parts = ['<', self.__class__.__name__, ' id=', str(self.id)]
        
        type_ = self.type
        if type_ is not IntegrationType.none:
            repr_parts.append(', type=')
            repr_parts.append(repr(type_))
        
        user = self.user
        if (user is not ZEROUSER):
            repr_parts.append(', user=')
            repr_parts.append(repr(user.full_name))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(DiscordEntity.__hash__)
    def __hash__(self):
        id_ = self.id
        if id_:
            return id_
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Hashes the fields of the integration.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # enabled
        hash_value ^= self.enabled << 2
        
        # name
        hash_value ^= hash(self.name)
        
        # metadata
        hash_value ^= hash(self.metadata)
        
        # type
        integration_type = self.type
        if (integration_type is not IntegrationType.none):
            hash_value ^= hash(integration_type)
        
        # user
        user = self.user
        if (user is not ZEROUSER):
            hash_value ^= hash(user)
        
        return hash_value
    

    def __eq__(self, other):
        """Returns whether the two integrations are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    

    def __ne__(self, other):
        """Returns whether the two integrations are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two integrations are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<cls>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id == other_id
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # metadata
        if self.metadata != other.metadata:
            return False
        
        # type
        if self.type != other.type:
            return False
        
        # user
        if self.user != other.user:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the integration.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = 0
        new.enabled = self.enabled
        new.name = self.name
        new.metadata = self.metadata.copy()
        new.type = self.type
        new.user = self.user
        return new
    
    
    def copy_with(
        self,
        *,
        integration_type = ...,
        enabled = ...,
        name = ...,
        user = ...,
        **keyword_parameters,
    ):
        """
        Copies the integration account with the given fields.
        
        Parameters
        ----------
        integration_type : `str`, ``IntegrationType``, Optional (Keyword only)
            The integration's type.
        
        enabled : `bool`, Optional (Keyword only)
            Whether this integration is enabled.
        
        name : `str`, Optional (Keyword only)
            The name of the integration.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            User for who the integration is.
        
        **keyword_parameters : Keyword parameters
            keyword parameters to set the integration's attributes as.
        
        Other Parameters
        ----------------
        account : ``IntegrationAccount``, ``ClientUserBase``, Optional (Keyword only)
            The integration's respective account. If the integration type is `'discord'`, then set as a discord user
            itself.
        
        application : ``IntegrationApplication``, Optional (Keyword only)
            The application of the integration if applicable.
        
        expire_behavior : `int`, ``IntegrationExpireBehavior``, Optional (Keyword only)
            The behavior of expiring subscription.
        
        expire_grace_period : `int`, Optional (Keyword only)
            The grace period in days for expiring subscribers.
        
        role_id : `int`, Optional (Keyword only)
            The role's identifier what the integration uses for subscribers.
        
        scopes : `None`, `iterable` of (`str`, ``Oauth2Scope``) items, Optional (Keyword only)
            The scopes the application was authorized with.
        
        subscriber_count : `int`, Optional (Keyword only)
            How many subscribers the integration has. 
        
        synced_at : `None`, `datetime`, Optional (Keyword only)
            When the integration was last synced.
        
        syncing : `bool`, Optional (Keyword only)
            Whether the integration syncing.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - Parameter type incorrect.
            - Extra or unused parameters.
        ValueError
            - Parameter value incorrect.
        """
        # integration_type
        if integration_type is ...:
            integration_type = self.type
        else:
            integration_type = validate_type(integration_type)
        
        # enabled
        if enabled is ...:
            enabled = self.enabled
        else:
            enabled = validate_enabled(enabled)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # metadata
        if integration_type is self.type:
            metadata = self.metadata.copy_with(keyword_parameters)
        else:
            metadata = integration_type.metadata_type(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        new = object.__new__(type(self))
        new.id = 0
        new.enabled = enabled
        new.name = name
        new.metadata = metadata
        new.type = integration_type
        new.user = user
        return new
    
    
    @property
    def detail(self):
        """
        ``Integration.detail`` is deprecated and will be removed in 2023 February. Please use ``.metadata`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.detail` is deprecated and will be removed in 2023 February. '
                f'Please use `.metadata` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.metadata
    
    
    @property
    def role(self):
        """
        Returns the integration's role.
        
        Returns
        -------
        role : `None`, ``Role``
        """
        role_id = self.role_id
        if role_id:
            return create_partial_role_from_id(role_id)
    
    # Metadata proxy properties
    
    @property
    @copy_docs(IntegrationMetadataBase.account)
    def account(self):
        return self.metadata.account
    
    
    @property
    @copy_docs(IntegrationMetadataBase.application)
    def application(self):
        return self.metadata.application
    
    
    @property
    @copy_docs(IntegrationMetadataBase.expire_behavior)
    def expire_behavior(self):
        return self.metadata.expire_behavior
    
    
    @property
    @copy_docs(IntegrationMetadataBase.expire_grace_period)
    def expire_grace_period(self):
        return self.metadata.expire_grace_period
    
    
    @property
    @copy_docs(IntegrationMetadataBase.revoked)
    def revoked(self):
        return self.metadata.revoked
    
    
    @property
    @copy_docs(IntegrationMetadataBase.role_id)
    def role_id(self):
        return self.metadata.role_id
    
    
    @property
    @copy_docs(IntegrationMetadataBase.scopes)
    def scopes(self):
        return self.metadata.scopes
    
    
    @property
    @copy_docs(IntegrationMetadataBase.subscriber_count)
    def subscriber_count(self):
        return self.metadata.subscriber_count
    
    
    @property
    @copy_docs(IntegrationMetadataBase.synced_at)
    def synced_at(self):
        return self.metadata.synced_at
    
    
    @property
    @copy_docs(IntegrationMetadataBase.syncing)
    def syncing(self):
        return self.metadata.syncing



__all__ = ()

from scarletio import Compound

from ...application import Application, EULA
from ...bases import maybe_snowflake
from ...channel import Channel
from ...core import EULAS
from ...embed import EmbedImage
from ...http import DiscordHTTPClient, is_media_url
from ...message import Attachment

from ..request_helpers import get_channel_id
from ..utils import Typer


def _assert__attachment(attachment):
    """
    Asserts whether the attachment's type is correct.
    
    Parameters
    ----------
    attachment : ``Attachment``, ``EmbedImage``
        The attachment object.
    
    Raises
    ------
    AssertionError
        - If `attachment` is not ``Attachment``, ``EmbedImage``.
    """
    if not isinstance(attachment, (Attachment, EmbedImage)):
        raise AssertionError(
            f'`attachment` can be `{Attachment.__name__}`, `{EmbedImage.__name__}`, got '
            f'{attachment.__class__.__name__}; {attachment!r}.'
        )
    
    return True


class ClientCompoundMiscellaneousEndpoints(Compound):
    
    http : DiscordHTTPClient
    

    async def download_attachment(self, attachment):
        """
        Downloads an attachment object's file. This method always prefers the proxy url of the attachment if applicable.
        
        This method is a coroutine.
        
        Parameters
        ----------
        attachment : ``Attachment``, ``EmbedImage``
            The attachment object what's file will be requested.
        
        Returns
        -------
        response_data : `bytes`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__attachment(attachment)
        
        url = attachment.proxy_url
        if (url is None) or is_media_url(url):
           url = attachment.url
        
        async with self.http.get(url) as response:
            return (await response.read())
    

    async def eula_get(self, eula):
        """
        Requests the eula with the given id.
        
        This method is a coroutine.
        
        Parameters
        ----------
        eula : `int`
            The `id` of the eula to request.

        Returns
        -------
        eula : ``EULA``
        
        Raises
        ------
        TypeError
            If `eula` was not given neither as ``EULA`` not as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(eula, EULA):
            eula_id = eula.id
        
        else:
            eula_id = maybe_snowflake(eula)
            if eula_id is None:
                raise TypeError(
                    f'`eula` can be `{EULA.__name__}`, `int`, got {eula.__class__.__name__}; {eula!r}.'
                )
            
            eula = EULAS.get(eula_id, None)
        
        
        eula_data = await self.http.eula_get(eula_id)
        if eula is None:
            eula = EULA.from_data(eula_data)
        else:
            eula._update_attributes(eula_data)
        
        return eula
    
    
    async def application_get_all_detectable(self):
        """
        Requests the detectable applications
        
        This method is a coroutine.
        
        Returns
        -------
        applications : `list` of ``Application``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        applications_data = await self.http.application_get_all_detectable()
        return [Application.from_data_detectable(application_data) for application_data in applications_data]
    
    
    async def typing(self, channel):
        """
        Sends a typing event to the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel where typing will be triggered.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        The client will be shown up as typing for 8 seconds, or till it sends a message at the respective channel.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_textual)
        
        await self.http.typing(channel_id)
    
    
    def keep_typing(self, channel, timeout = 300.0):
        """
        Returns a context manager which will keep sending typing events at the channel. Can be used to indicate that
        the bot is working.
        
        Parameters
        ----------
        channel ``Channel``, `int`
            The channel where typing will be triggered.
        timeout : `float` = `300.0`, Optional
            The maximal duration for the ``Typer`` to keep typing.
        
        Returns
        -------
        typer : ``Typer``
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        
        Examples
        --------
        ```py
        with client.keep_typing(channel):
            # Do some things
            await client.message_create(channel, 'Ayaya')
        ```
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_textual)
        
        return Typer(self, channel_id, timeout)



__all__ = ()

import re

from scarletio import Compound

from ...bases import maybe_snowflake, maybe_snowflake_pair
from ...core import EMOJIS, GUILDS
from ...emoji import Emoji
from ...http import DiscordHTTPClient
from ...role import Role
from ...utils import image_to_base64
from ..request_helpers import get_guild_and_id, get_guild_id, get_emoji_guild_id_and_id


_VALID_NAME_CHARS = re.compile('([0-9A-Za-z_]+)')


class ClientCompoundEmojiEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def emoji_get(self, emoji, force_update = False):
        """
        Requests the emoji by it's id at the given guild. If the client's logging in is finished, then it should have
        it's every emoji loaded already.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `tuple` (`int`, `int`) items
            The emoji, or 2 snowflake representing it.
        
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the emoji should be requested even if it supposed to be up to date.
        
        Returns
        -------
        emoji : ``Emoji``
        
        Raises
        ------
        TypeError
            If `emoji` was not given as ``Emoji``, `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(emoji, Emoji):
            guild_id = emoji.guild_id
            emoji_id = emoji.id
        else:
            snowflake_pair = maybe_snowflake_pair(emoji)
            if snowflake_pair is None:
                raise TypeError(
                    f'`emoji` can be `{Emoji.__name__}`, `tuple` (`int`, `int`), '
                    f'got {emoji.__class__.__name__}; {emoji!r}.'
                )
            
            guild_id, emoji_id = snowflake_pair
            emoji = EMOJIS.get(emoji_id, None)
        
        
        # If the emoji has no linked guild, we cannot request it, so we return instantly.
        if not guild_id:
            # Create empty emoji instead of returning `None` to stop expected errors
            if emoji is None:
                emoji = Emoji.precreate(emoji_id)
            
            return emoji
        
        if (emoji is not None) and (not emoji.partial) and (not force_update):
            return emoji
        
        emoji_data = await self.http.emoji_get(guild_id, emoji_id)
        
        if (emoji is None):
            emoji = Emoji(emoji_data, guild_id)
        else:
            emoji._set_attributes(emoji_data, guild_id)
        
        return emoji
    
    
    async def emoji_guild_get_all(self, guild):
        """
        Syncs the given guild's emojis with the wrapper.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's emojis will be synced.
        
        Returns
        -------
        emojis : `list` of ``Emoji``
            The guild's emojis.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        emoji_datas = await self.http.emoji_guild_get_all(guild_id)
        
        if guild is None:
            # Do not create a partial guild, because it would have been garbage collected after leaving the function
            # anyways
            guild = GUILDS.get(guild_id, None)
        
        if guild is None:
            emojis = [Emoji(emoji_data, guild_id) for emoji_data in emoji_datas]
        else:
            guild._sync_emojis(emoji_datas)
            emojis = list(guild.emojis.values())
        
        return emojis
    
    
    async def emoji_create(self, guild, name, image, *, roles=None, reason = None):
        """
        Creates an emoji at the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, where the emoji will be created.
        name : `str`
            The emoji's name. It's length can be between `2` and `32`.
        image : `bytes-like`
            The emoji's icon.
        roles : `None`, (`list`, `set`, `tuple`) of (``Role``, `int`) = `None`, Optional (Keyword only)
            Whether the created emoji should be limited only to users with any of the specified roles.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the guild's audit logs.
        
        Returns
        -------
        emoji : ``Emoji``
            The created emoji.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
            If `roles` contains a non ``Role``, `int` element.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was not given as `str`.
            - If `name` length is out of the expected range [1:32].
            - If `roles` was not given neither as `None`, `list`, `tuple`, `set`.
        Notes
        -----
        Only some characters can be in the emoji's name, so every other character is filtered out.
        """
        guild_id = get_guild_id(guild)
        
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                )
        
        name = ''.join(_VALID_NAME_CHARS.findall(name))
        
        if __debug__:
            name_length = len(name)
            if name_length < 2 or name_length > 32:
                raise AssertionError(
                    f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                )
        
        role_ids = set()
        if (roles is not None):
            if __debug__:
                if not isinstance(roles, (list, set, tuple)):
                    raise AssertionError(
                        f'`roles` can be `None`, `list`, `set`, `tuple`, got {roles.__class__.__name__}; {roles!r}.'
                    )
            
            for role in roles:
                if isinstance(role, Role):
                    role_id = role.id
                else:
                    role_id = maybe_snowflake(role)
                    if role_id is None:
                        raise TypeError(
                            f'`roles` contains not `{Role.__name__}`, `int` element, '
                            f'got {role.__class__.__name__}; {role!r}; roles={roles!r}.'
                        )
                
                role_ids.add(role_id)
            
        image = image_to_base64(image)
        
        data = {
            'name': name,
            'image': image,
            'roles': role_ids
        }
        
        data = await self.http.emoji_create(guild_id, data, reason)
        
        emoji = Emoji(data, guild_id)
        emoji.user = self
        return emoji
    
    
    async def emoji_delete(self, emoji, *, reason = None):
        """
        Deletes the given emoji.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `emoji` was not given as ``Emoji``.
        """
        guild_id, emoji_id = get_emoji_guild_id_and_id(emoji)
        
        # Cannot delete partially loaded emoji.
        if guild_id is None:
            return
        
        await self.http.emoji_delete(guild_id, emoji_id, reason = reason)
    
    
    async def emoji_edit(self, emoji, *, name=..., roles=..., reason = None):
        """
        Edits the given emoji.
        
        This method is a coroutine.
        
        Parameters
        ----------
        emoji : ``Emoji``, `tuple` (`int`, `int`)
            The emoji to edit.
        name : `str`, Optional (Keyword only)
            The emoji's new name. It's length can be in range [2:32].
        roles : `None`, (`list`, `set`, `tuple`) of (``Role``, `int`), Optional (Keyword only)
            The roles to what is the role limited. By passing it as `None`, or as an empty `list` you can remove the
            current ones.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `roles` contains a non ``Role``, `int` element.
            - If `emoji` is not ``Emoji``, `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was not given as `str`.
            - If `name` length is out of the expected range [1:32].
            - If `roles` was not given neither as `None`, `list`, `tuple`, `set`.
        """
        guild_id, emoji_id = get_emoji_guild_id_and_id(emoji)
        
        # Cannot edit partially loaded emojis.
        if not guild_id:
            return
        
        data = {}
        
        # name is not a required parameter anymore, we can edit the emoji normally.
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                    )
            
            name = ''.join(_VALID_NAME_CHARS.findall(name))
            
            if __debug__:
                name_length = len(name)
                if name_length < 2 or name_length > 32:
                    raise AssertionError(
                        f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                    )
            
            data['name'] = name
        
        
        if (roles is not ...):
            role_ids = set()
            if (roles is not None):
                if __debug__:
                    if not isinstance(roles, (list, set, tuple)):
                        raise AssertionError(
                            f'`roles` can be `None`, `list`, `set`, `tuple`, got '
                            f'{roles.__class__.__name__}; {roles!r}.'
                        )
                
                for role in roles:
                    if isinstance(role, Role):
                        role_id = role.id
                    else:
                        role_id = maybe_snowflake(role)
                        if role_id is None:
                            raise TypeError(
                                f'`roles` contains not `{Role.__name__}`, `int` element, got '
                                f'{role.__class__.__name__}; {role!r}; roles={roles!r}.'
                            )
                    
                    role_ids.add(role_id)
            
            data['roles'] = role_ids
        
        
        await self.http.emoji_edit(guild_id, emoji_id, data, reason)



__all__ = ('IntegrationMetadataSubscription',)

from scarletio import copy_docs

from ...core import ROLES

from ..integration_account import IntegrationAccount

from .constants import EXPIRE_GRACE_PERIOD_DEFAULT, SUBSCRIBER_COUNT_DEFAULT
from .fields import (
    parse_expire_behavior, parse_expire_grace_period, parse_revoked, parse_role_id, parse_subscriber_count,
    parse_synced_at, parse_syncing, put_expire_behavior_into, put_expire_grace_period_into, put_revoked_into,
    put_role_id_into, put_subscriber_count_into, put_synced_at_into, put_syncing_into, validate_expire_behavior,
    validate_expire_grace_period, validate_revoked, validate_role_id, validate_subscriber_count, validate_synced_at,
    validate_syncing, validate_account, parse_account, put_account_into
)
from .preinstanced import IntegrationExpireBehavior

from .base import IntegrationMetadataBase


class IntegrationMetadataSubscription(IntegrationMetadataBase):
    """
    Metadata of a linked subscription integration.
    
    Attributes
    ----------
    account : ``IntegrationAccount``
        The integration's respective account.
    expire_behavior : ``IntegrationExpireBehavior``
        The behavior of expiring subscription.
    expire_grace_period : `int`
        The grace period in days for expiring subscribers.
    revoked : `bool`
        Whether the integration is removed.
    role_id : `int`
        The role's identifier what the integration uses for subscribers.
    subscriber_count : `int`
        How many subscribers the integration has.
    synced_at : `None`, `datetime`
        When the integration was last synced.
    syncing : `bool`
        Whether the integration syncing.
    """
    __slots__ = (
        'account', 'expire_behavior', 'expire_grace_period', 'revoked', 'role_id', 'subscriber_count', 'synced_at',
        'syncing'
    )
    
    @copy_docs(IntegrationMetadataBase.__new__)
    def __new__(cls, keyword_parameters):
        # account
        try:
            account = keyword_parameters.pop('account')
        except KeyError:
            account = IntegrationAccount._create_empty()
        else:
            account = validate_account(account)
        
        # expire_behavior
        try:
            expire_behavior = keyword_parameters.pop('expire_behavior')
        except KeyError:
            expire_behavior = IntegrationExpireBehavior.remove_role
        else:
            expire_behavior = validate_expire_behavior(expire_behavior)
        
        # expire_grace_period
        try:
            expire_grace_period = keyword_parameters.pop('expire_grace_period')
        except KeyError:
            expire_grace_period = EXPIRE_GRACE_PERIOD_DEFAULT
        else:
            expire_grace_period = validate_expire_grace_period(expire_grace_period)
        
        # revoked
        try:
            revoked = keyword_parameters.pop('revoked')
        except KeyError:
            revoked = False
        else:
            revoked = validate_revoked(revoked)
        
        # role_id
        try:
            role_id = keyword_parameters.pop('role_id')
        except KeyError:
            role_id = 0
        else:
            role_id = validate_role_id(role_id)
        
        # subscriber_count
        try:
            subscriber_count = keyword_parameters.pop('subscriber_count')
        except KeyError:
            subscriber_count = SUBSCRIBER_COUNT_DEFAULT
        else:
            subscriber_count = validate_subscriber_count(subscriber_count)
        
        # synced_at
        try:
            synced_at = keyword_parameters.pop('synced_at')
        except KeyError:
            synced_at = None
        else:
            synced_at = validate_synced_at(synced_at)
        
        # syncing
        try:
            syncing = keyword_parameters.pop('syncing')
        except KeyError:
            syncing = False
        else:
            syncing = validate_syncing(syncing)
        
        self = object.__new__(cls)
        self.account = account
        self.expire_behavior = expire_behavior
        self.expire_grace_period = expire_grace_period
        self.revoked = revoked
        self.role_id = role_id
        self.subscriber_count = subscriber_count
        self.synced_at = synced_at
        self.syncing = syncing
        return self
    
    
    @classmethod
    def from_role(cls, role_id):
        """
        Creates a linked integration metadata.
        
        Parameters
        ----------
        role_id : ``Role``
            The respective role's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = cls._create_empty()
        self.role_id = role_id
        return self
        
    
    @classmethod
    @copy_docs(IntegrationMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.account = parse_account(data)
        self.expire_behavior = parse_expire_behavior(data)
        self.expire_grace_period = parse_expire_grace_period(data)
        self.revoked = parse_revoked(data)
        self.role_id = parse_role_id(data)
        self.subscriber_count = parse_subscriber_count(data)
        self.synced_at = parse_synced_at(data)
        self.syncing = parse_syncing(data)
        return self
    
    
    @copy_docs(IntegrationMetadataBase.to_data)
    def to_data(self, defaults = False):
        data = {}
        
        put_account_into(self.account, data, defaults)
        put_expire_behavior_into(self.expire_behavior, data, defaults)
        put_expire_grace_period_into(self.expire_grace_period, data, defaults)
        put_revoked_into(self.revoked, data, defaults)
        put_role_id_into(self.role_id, data, defaults)
        put_subscriber_count_into(self.subscriber_count, data, defaults)
        put_synced_at_into(self.synced_at, data, defaults)
        put_syncing_into(self.syncing, data, defaults)
        
        return data
    
    
    @copy_docs(IntegrationMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<', self.__class__.__name__,
        ]
        
        role_id = self.role_id
        if role_id:
            try:
                role = ROLES[role_id]
            except KeyError:
                pass
            else:
                repr_parts.append(' role=')
                repr_parts.append(repr(role))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(IntegrationMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.account != other.account:
            return False
        
        if self.expire_behavior != other.expire_behavior:
            return False
        
        if self.expire_grace_period != other.expire_grace_period:
            return False
        
        if self.revoked != other.revoked:
            return False
        
        if self.role_id != other.role_id:
            return False
        
        if self.subscriber_count != other.subscriber_count:
            return False
        
        if self.synced_at != other.synced_at:
            return False
        
        if self.syncing != other.syncing:
            return False
        
        return True
    
    
    @copy_docs(IntegrationMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # account
        hash_value ^= hash(self.account)
        
        # expire_behavior
        hash_value ^= self.expire_behavior.value
        
        # expire_grace_period
        hash_value ^= self.expire_grace_period << 4
        
        # revoked
        hash_value ^= self.revoked << 8
        
        # role_id
        hash_value ^= self.role_id
        
        # subscriber_count
        hash_value ^= self.subscriber_count << 12
        
        # synced_at
        synced_at = self.synced_at
        if (synced_at is not None):
            hash_value ^= hash(synced_at)
        
        # syncing
        hash_value ^= self.syncing << 20
        
        return hash_value
    
    
    @classmethod
    @copy_docs(IntegrationMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.account = IntegrationAccount._create_empty()
        self.expire_behavior = IntegrationExpireBehavior.remove_role
        self.expire_grace_period = EXPIRE_GRACE_PERIOD_DEFAULT
        self.revoked = False
        self.role_id = 0
        self.subscriber_count = SUBSCRIBER_COUNT_DEFAULT
        self.synced_at = None
        self.syncing = False
        return self
    
    
    @copy_docs(IntegrationMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.account = self.account
        new.expire_behavior = self.expire_behavior
        new.expire_grace_period = self.expire_grace_period
        new.revoked = self.revoked
        new.role_id = self.role_id
        new.subscriber_count = self.subscriber_count
        new.synced_at = self.synced_at
        new.syncing = self.syncing
        return new
    
    
    @copy_docs(IntegrationMetadataBase.copy_with)
    def copy_with(self, keyword_parameters):
        # account
        try:
            account = keyword_parameters.pop('account')
        except KeyError:
            account = self.account
        else:
            account = validate_account(account)
        
        # expire_behavior
        try:
            expire_behavior = keyword_parameters.pop('expire_behavior')
        except KeyError:
            expire_behavior = self.expire_behavior
        else:
            expire_behavior = validate_expire_behavior(expire_behavior)
        
        # expire_grace_period
        try:
            expire_grace_period = keyword_parameters.pop('expire_grace_period')
        except KeyError:
            expire_grace_period = self.expire_grace_period
        else:
            expire_grace_period = validate_expire_grace_period(expire_grace_period)
        
        # revoked
        try:
            revoked = keyword_parameters.pop('revoked')
        except KeyError:
            revoked = self.revoked
        else:
            revoked = validate_revoked(revoked)
        
        # role_id
        try:
            role_id = keyword_parameters.pop('role_id')
        except KeyError:
            role_id = self.role_id
        else:
            role_id = validate_role_id(role_id)
        
        # subscriber_count
        try:
            subscriber_count = keyword_parameters.pop('subscriber_count')
        except KeyError:
            subscriber_count = self.subscriber_count
        else:
            subscriber_count = validate_subscriber_count(subscriber_count)
        
        # synced_at
        try:
            synced_at = keyword_parameters.pop('synced_at')
        except KeyError:
            synced_at = self.synced_at
        else:
            synced_at = validate_synced_at(synced_at)
        
        # syncing
        try:
            syncing = keyword_parameters.pop('syncing')
        except KeyError:
            syncing = self.syncing
        else:
            syncing = validate_syncing(syncing)
        
        new = object.__new__(type(self))
        new.account = account
        new.expire_behavior = expire_behavior
        new.expire_grace_period = expire_grace_period
        new.revoked = revoked
        new.role_id = role_id
        new.subscriber_count = subscriber_count
        new.synced_at = synced_at
        new.syncing = syncing
        return new



from .base import *
from .constants import *
from .discord import *
from .fields import *
from .preinstanced import *
from .subscription import *


__all__ = (
    *base.__all__,
    *constants.__all__,
    *discord.__all__,
    *fields.__all__,
    *preinstanced.__all__,
    *subscription.__all__,
)



__all__ = ('unload_plugin',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.unload)
def unload_plugin(name, *, blocking=True, deep=True):
    return PLUGIN_LOADER.unload(name, blocking=blocking, deep=deep)



__all__ = ()

from scarletio import Compound

from ...channel import Channel
from ...http import DiscordHTTPClient
from ...scheduled_event import PrivacyLevel
from ...stage import Stage

from ..request_helpers import get_channel_id, get_stage_channel_id


class ClientCompoundStageEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def stage_create(
        self, channel, topic, *, privacy_level=PrivacyLevel.guild_only, send_start_notification=False,
    ):
        """
        Edits the given stage channel.
        
        Will trigger a `stage_create` event.
        
        > The endpoint has a long rate limit, so please use ``.stage_edit`` to edit just the stage's topic.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to edit.
        
        topic : `None`, `str`
            The new topic of the stage.
        
        privacy_level : ``PrivacyLevel``, `int` = `PrivacyLevel.guild_only`, Optional (Keyword only)
            The new privacy level of the stage. Defaults to guild only.
        
        send_start_notification : `bool` = `False`, Optional (Keyword only)
            Whether @everyone should be notified when the stage is started.
            
            > You must have `mention everyone` permission.
        
        Returns
        -------
        stage : ``Stage``
            The created stage instance.
        
        Raises
        ------
        TypeError
            - If `channel` was not given as ``Channel`` neither as `int`.
            - If `privacy_level` was not given neither as ``PrivacyLevel`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `topic` was not given neither as `None`, `str`.
            - If `topic`'s length is out of range [1:120].
            - If `send_start_notification` is not `bool`
        """
        # channel_id
        channel_id = get_channel_id(channel, Channel.is_guild_stage)
        
        # topic
        if topic is None:
            topic = ''
        else:
            if __debug__:
                if not isinstance(topic, str):
                    raise AssertionError(
                        f'`topic` can be `None`, `str`, got {topic.__class__.__name__}; {topic!r}.'
                    )
                
                topic_length = len(topic)
                if (topic_length < 1) or (topic_length > 120):
                    raise AssertionError(
                        f'`topic` length can be in range [1:120], got {topic_length!r}; {topic!r}.'
                    )
        
        # privacy_level
        if isinstance(privacy_level, PrivacyLevel):
            privacy_level = privacy_level.value
        
        elif isinstance(privacy_level, int):
            privacy_level = privacy_level
        
        else:
            raise TypeError(
                f'`privacy_level` can be `{PrivacyLevel.__name__}`, `int` , got '
                f'{privacy_level.__class__.__name__}; {privacy_level!r}.'
            )
        
        # send_start_notification
        if __debug__:
            if not isinstance(send_start_notification, bool):
                raise AssertionError(
                    f'`send_start_notification` can be `bool`, '
                    f'got {send_start_notification.__class__.__name__}; {send_start_notification!r}.'
                )
        
        
        data = {
            'channel_id': channel_id,
            'topic': topic,
            'privacy_level': privacy_level,
        }
        
        if send_start_notification:
            data['send_start_notification'] = True
        
        
        data = await self.http.stage_create(data)
        return Stage(data)
    
    
    async def stage_edit(self, stage, topic=..., *, privacy_level=...):
        """
        Edits the given stage channel.
        
        Will trigger a `stage_edit` event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        stage : ``Stage``, ``Channel``, `int`
            The stage to edit. Can be given as it's channel's identifier.
        topic : `str`
            The new topic of the stage.
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The new privacy level of the stage.
        
        Raises
        ------
        TypeError
            - If `stage` was not given as ``Stage``, ``Channel`` neither as `int`.
            - If `privacy_level` was not given neither as ``PrivacyLevel`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `topic` was not given neither as `None` nor as `str`.
            - If `topic`'s length is out of range [1:120].
        """
        channel_id = get_stage_channel_id(stage)
        
        data = {}
        
        if (topic is not ...):
            if __debug__:
                if not isinstance(topic, str):
                    raise AssertionError(
                        f'`topic` can be `None`, `str`, got {topic.__class__.__name__}; {topic!r}.'
                    )
                
                topic_length = len(topic)
                if (topic_length < 1) or (topic_length > 120):
                    raise AssertionError(
                        f'`topic` length can be in range [1:120], got {topic_length!r}; {topic!r}.'
                    )
            
            data['topic'] = topic
        
        
        if (privacy_level is not ...):
            if isinstance(privacy_level, PrivacyLevel):
                privacy_level = privacy_level.value
            elif isinstance(privacy_level, int):
                privacy_level = privacy_level
            else:
                raise TypeError(
                    f'`privacy_level` can be `{PrivacyLevel.__name__}`, `int` , got '
                    f'{privacy_level.__class__.__name__}; {privacy_level!r}.'
                )
            
            data['privacy_level'] = privacy_level
        
        
        if not data:
            return
        
        await self.http.stage_edit(channel_id, data)
        # We receive data, but ignore it, so we can dispatch it.
    
    
    async def stage_delete(self, stage):
        """
        Deletes the given stage channel.
        
        Will trigger a `stage_delete` event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        stage : ``Stage``, ``Channel``, `int`
            The stage to delete. Can be given as it's channel's identifier.
        
        Raises
        ------
        TypeError
            If `stage` was not given as ``Stage``, ``Channel`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_stage_channel_id(stage)
        
        await self.http.stage_delete(channel_id)
        # We receive no data.
    
    
    async def stage_get(self, channel):
        """
        Gets the stage of the given stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The stage's channel's identifier.
        
        Raises
        ------
        TypeError
            If `channel` was not given as ``Channel`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_guild_stage)
        
        data = await self.http.stage_get(channel_id)
        return Stage(data)



from .alert_message import *
from .base import *
from .constants import *
from .fields import *
from .timeout import *


__all__ = (
    *alert_message.__all__,
    *base.__all__,
    *constants.__all__,
    *fields.__all__,
    *timeout.__all__,
)
